#!/usr/bin/env python

# Version 2.1.1
#
# Author: Benjamin Cance (bjc@tdx.li)
# Copyright Benjamin Cance 2024
#
# 2-Aug-24 
# - Updating to current PEP

VERSION='2.1.1'

import sys
import csv
import logging
from pathlib import Path
from typing import TextIO
from .mft_analyzer import MFTAnalyzer
from .mft_formatters import mft_to_csv, mft_to_body, mft_to_l2t

class MftSession:
    def __init__(self):
        self.mft = {}
        self.folders = {}
        self.options = None
        self.file_mft: TextIO = None
        self.file_csv = None
        self.file_body: TextIO = None
        self.file_csv_time: TextIO = None
        self.setup_logging()

    def setup_logging(self):
        logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

    def open_files(self):
        try:
            self.file_mft = open(self.options.filename, 'rb')
            if not self.file_mft:
                parser.error("The --filename argument is required")
        except IOError as e:
            logging.error(f"Unable to open file: {self.options.filename}. Error: {e}")
            sys.exit(1)

        if self.options.output:
            try:
                with open(self.options.output, 'w', newline='') as output_file:
                    self.file_csv = csv.writer(output_file, dialect=csv.excel, quoting=csv.QUOTE_ALL)
            except IOError as e:
                logging.error(f"Unable to open file: {self.options.output}. Error: {e}")
                sys.exit(1)

        if self.options.bodyfile:
            try:
                self.file_body = open(self.options.bodyfile, 'w')
            except IOError as e:
                logging.error(f"Unable to open file: {self.options.bodyfile}. Error: {e}")
                sys.exit(1)

        if self.options.csvtimefile:
            try:
                self.file_csv_time = open(self.options.csvtimefile, 'w')
            except IOError as e:
                logging.error(f"Unable to open file: {self.options.csvtimefile}. Error: {e}")
                sys.exit(1)

    def process_mft_file(self):
        analyzer = MFTAnalyzer(self.options)
        analyzer.process_mft_file(self.file_mft)
        self.mft = analyzer.mft
        self.folders = analyzer.folders

    def print_records(self):
        for i, record in self.mft.items():
            if self.file_csv:
                self.file_csv.writerow(mft_to_csv(record, False))
            if self.file_csv_time:
                self.file_csv_time.write(mft_to_l2t(record))
            if self.file_body:
                self.file_body.write(mft_to_body(record, self.options.bodyfull, self.options.bodystd))

    def close_files(self):
        for file in [self.file_mft, self.file_body, self.file_csv_time]:
            if file:
                file.close()
        if self.file_csv:
            self.file_csv.close()

    def run(self):
        try:
            self.open_files()
            self.process_mft_file()
            self.print_records()
        finally:
            self.close_files()



# Version 2.1.1
#
# Author: Benjamin Cance (bjc@tdx.li)
#
# 2-Aug-24 
# - Updating to current PEP

import sys
import logging
from analyzemft import mft_session
from analyzemft.mft import set_default_options

def main():
    # Set up logging
    logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
    
    try:
        # Get the command line options
        parser = set_default_options()
        options = parser.parse_args()
        
        # Create and initialize the MftSession
        session = mft_session.MftSession()
        session.options = options

        # Set up debug logging if requested
        if options.debug:
            logging.getLogger().setLevel(logging.DEBUG)

        # Run the MFT analysis
        session.open_files()
        session.process_mft_file()
        session.print_records()
        session.close_files()

    except Exception as e:
        logging.error(f"An error occurred: {e}")
        sys.exit(1)

if __name__ == "__main__":
    main()



import argparse
import logging
from typing import Dict, Any
from . import mft_utils
from .mft_formatters import mft_to_csv, mft_to_body, mft_to_l2t, mft_to_json
from .mft_utils import decodeMFTmagic, decodeMFTisactive, decodeMFTrecordtype, decodeVolumeInfo, decodeObjectID, ObjectID

def set_default_options() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="MFT Analyzer")
    parser.add_argument("--debug", action="store_true", default=False, help="Enable debug logging")
    parser.add_argument("--localtz", default=None, help="Use local timezone")
    parser.add_argument("--bodystd", action="store_true", default=False, help="Use standard body format")
    parser.add_argument("--bodyfull", action="store_true", default=False, help="Use full body format")
    parser.add_argument("--json", action="store_true", default=False, help="Output in JSON format")
    return parser

def setup_logging(debug: bool):
    level = logging.DEBUG if debug else logging.INFO
    logging.basicConfig(level=level, format='%(asctime)s - %(levelname)s - %(message)s')

def read_mft_file(filename: str) -> bytes:
    try:
        with open(filename, 'rb') as file:
            return file.read(1024)
    except IOError as e:
        logging.error(f"Error reading MFT file: {e}")
        raise



#!/usr/bin/env python

# Version 2.1.1
#
# Author: Benjamin Cance (bjc@tdx.li)
# Copyright Benjamin Cance 2024
#
# 2-Aug-24 
# - Updating to current PEP
# - Correctly calculate the Unix timestamp from Windows filetime

from datetime import datetime, timezone
from typing import Union, Dict, Any
import struct
import logging
from enum import Enum

class WindowsTime:
    def __init__(self, low: int, high: int, localtz: bool):
        self.low = int(low)
        self.high = int(high)
        self.dt: Union[datetime, int] = 0
        self.dtstr: str = ""
        self.unixtime: float = 0.0

        if (low == 0) and (high == 0):
            self.dtstr = "Not defined"
            return

        try:
            self.unixtime = self.get_unix_time()
            self.dt = datetime.fromtimestamp(self.unixtime, tz=timezone.utc if not localtz else None)
            self.dtstr = self.dt.isoformat(' ')

        except (OverflowError, OSError) as e:
            logging.error(f"Error converting timestamp: {e}")
            self.dtstr = "Invalid timestamp"
            self.unixtime = 0.0

    def get_unix_time(self) -> float:
        wintime = (self.high << 32) | self.low
        return wintime / 10000000 - 11644473600

    def __str__(self):
        return self.dtstr

class MagicValues(Enum):
    GOOD = 0x454c4946
    BAD = 0x44414142
    ZERO = 0x00000000

def decodeMFTmagic(record: Dict[str, Any]) -> str:
    magic = record.get('magic', 0)
    return MagicValues(magic).name if magic in MagicValues._value2member_map_ else 'Unknown'

def decodeMFTisactive(record: Dict[str, Any]) -> str:
    return 'Active' if record.get('flags', 0) & 0x0001 else 'Inactive'

def decodeMFTrecordtype(record: Dict[str, Any]) -> str:
    flags = int(record.get('flags', 0))
    record_type = 'Folder' if flags & 0x0002 else 'File'
    if flags & 0x0004:
        record_type += ' + Unknown1'
    if flags & 0x0008:
        record_type += ' + Unknown2'
    return record_type

def decodeVolumeInfo(s,options):

    d = {}
    d['f1'] = struct.unpack("<d",s[:8])[0]                  # 8
    d['maj_ver'] = struct.unpack("B",s[8])[0]               # 1
    d['min_ver'] = struct.unpack("B",s[9])[0]               # 1
    d['flags'] = struct.unpack("<H",s[10:12])[0]            # 2
    d['f2'] = struct.unpack("<I",s[12:16])[0]               # 4

    if options.debug:
        print(f"+Volume Info")
        print(f"++F1%d" % d['f1'])
        print(f"++Major Version: %d" % d['maj_ver'])
        print(f"++Minor Version: %d" % d['min_ver'])
        print(f"++Flags: %d" % d['flags'])
        print(f"++F2: %d" % d['f2'])

    return d

def decodeObjectID(s):

    d = {}
    d['objid'] = ObjectID(s[0:16])
    d['orig_volid'] = ObjectID(s[16:32])
    d['orig_objid'] = ObjectID(s[32:48])
    d['orig_domid'] = ObjectID(s[48:64])

    return d

def ObjectID(s: bytes) -> str:
    if s == b'\x00' * 16:
        return 'Undefined'
    return f"{s[:4].hex()}-{s[4:6].hex()}-{s[6:8].hex()}-{s[8:10].hex()}-{s[10:16].hex()}"




#!/usr/bin/env python

# Version 2.1.1
#
# Author: Benjamin Cance (bjc@tdx.li)
# Copyright Benjamin Cance 2024
#
# 2-Aug-24 
# - Updating to current PEP


from . import mft_utils
from . import mft
from . import mft_session

__all__ = ["mft_utils", "mft", "mft_session"]


#!/usr/bin/env python

# Version 2.1.1
# Author: Benjamin Cance (bjc@tdx.li)
# Copyright Benjamin Cance 2024

import struct
import logging
from typing import Dict, Any, Callable
from .mft_utils import WindowsTime

class MFTAnalyzer:
    def __init__(self, options: Any):
        self.options = options
        self.mft: Dict[int, Dict[str, Any]] = {}
        self.folders: Dict[str, str] = {}
        self.num_records: int = 0
        self.attribute_handlers: Dict[int, Callable] = {
            0x10:  self.handle_standard_information,
            0x20:  self.handle_attribute_list,
            0x30:  self.handle_file_name,
            0x40:  self.handle_object_id,
            0x50:  self.handle_security_descriptor,
            0x60:  self.handle_volume_name,
            0x70:  self.handle_volume_information,
            0x80:  self.handle_data,
            0x90:  self.handle_index_root,
            0xA0:  self.handle_index_allocation,
            0xB0:  self.handle_bitmap,
            0xC0:  self.handle_reparse_point,
            0xD0:  self.handle_ea_information,
            0xE0:  self.handle_ea,
            0xF0:  self.handle_property_set,
            0x100: self.handle_logged_utility_stream,
        }
        self.setup_logging()
    def setup_logging(self):
        level = logging.DEBUG if self.options.debug else logging.INFO
        logging.basicConfig(level=level, format='%(asctime)s - %(levelname)s - %(message)s')

    def decode_unicode(self, s: bytes, length: int) -> str:
        # More effective handling of unicode 
        try:
            # Try UTF-16-LE first (standard for NTFS)
            return s[:length*2].decode('utf-16-le')
        except UnicodeDecodeError:
            try:
                # If UTF-16-LE fails, try UTF-8
                return s[:length].decode('utf-8')
        
            except UnicodeDecodeError:
                logging.warning(f"Failed to decode: {s[:length].hex()}")
                return ' '.join([f'0x{b:02x}' for b in s[:length]])

    def process_mft_file(self, file_mft):
        self.num_records = 0
        while True:
            raw_record = file_mft.read(1024)
            if not raw_record:
                break
            try:
                record = self.parse_record(raw_record)
                self.mft[self.num_records] = record
                self.num_records += 1
                if self.num_records % 1000 == 0:
                    logging.info(f"Processed {self.num_records} records")
            except Exception as e:
                logging.error(f"Error processing record {self.num_records}: {e}")

        self.gen_filepaths()

    def parse_record(self, raw_record: bytes) -> Dict[str, Any]:
        record: Dict[str, Any] = {'filename': '', 'notes': '', 'fncnt': 0}

        self.decodeMFTHeader(record, raw_record)

        record_number = record['recordnum']

        if self.options.debug:
            print(f"-->Record number: {record_number}\n\tMagic: {record['magic']} Attribute offset: {record['attr_off']} Flags: {hex(int(record['flags']))} Size:{record['size']}")

        if record['magic'] == 0x44414142:
            if self.options.debug:
                print("BAAD MFT Record")
            record['baad'] = True
            return record
        

        if record['magic'] != 0x454c4946:
            if self.options.debug:
                print("Corrupt MFT Record")
            record['corrupt'] = True
            return record

        read_ptr = record['attr_off']

        while read_ptr < 1024:
            ATRrecord = self.decodeATRHeader(raw_record[read_ptr:])
            if ATRrecord['type'] == 0xffffffff:  
                break

            if self.options.debug:
                print(f"Attribute type: {ATRrecord['type']:x} Length: {ATRrecord['len']} Res: {ATRrecord['res']:x}")

            handler = self.attribute_handlers.get(ATRrecord['type'], self.handle_unknown_attribute)
            handler(self, ATRrecord, raw_record[read_ptr:], record)

            if ATRrecord['len'] > 0:
                read_ptr += ATRrecord['len']
            else:
                if self.options.debug:
                    print("ATRrecord->len <= 0, exiting loop")
                break

        return record

    def gen_filepaths(self):
        for i in self.mft:
            if self.mft[i]['filename'] == '':
                if self.mft[i]['fncnt'] > 0:
                    self.get_folder_path(i)
                    if self.options.debug:
                        print(f"Filename (with path): {self.mft[i]['filename']}")
                else:
                    self.mft[i]['filename'] = 'NoFNRecord'

    def get_folder_path(self, seqnum: int) -> str:
        if self.options.debug:
            print(f"Building Folder For Record Number ({seqnum})")

        if seqnum not in self.mft:
            return 'Orphan'

        if self.mft[seqnum]['filename'] != '':
            return self.mft[seqnum]['filename']

        try:
            if self.mft[seqnum]['fn', 0]['par_ref'] == 5:
                self.mft[seqnum]['filename'] = '/' + self.mft[seqnum]['fn', self.mft[seqnum]['fncnt']-1]['name']
                return self.mft[seqnum]['filename']
        except:
            self.mft[seqnum]['filename'] = 'NoFNRecord'
            return self.mft[seqnum]['filename']

        if self.mft[seqnum]['fn', 0]['par_ref'] == seqnum:
            if self.options.debug:
                print(f"Error, self-referential, while trying to determine path for seqnum {seqnum}")
            self.mft[seqnum]['filename'] = 'ORPHAN/' + self.mft[seqnum]['fn', self.mft[seqnum]['fncnt']-1]['name']
            return self.mft[seqnum]['filename']

        parentpath = self.get_folder_path(self.mft[seqnum]['fn', 0]['par_ref'])
        self.mft[seqnum]['filename'] = parentpath + '/' + self.mft[seqnum]['fn', self.mft[seqnum]['fncnt']-1]['name']

        return self.mft[seqnum]['filename']
    
    def decodeMFTHeader(self, record: Dict[str, Any], raw_record: bytes) -> None:
        record['magic'] = struct.unpack("<I", raw_record[:4])[0]
        record['upd_off'] = struct.unpack("<H", raw_record[4:6])[0]
        record['upd_cnt'] = struct.unpack("<H", raw_record[6:8])[0]
        record['lsn'] = struct.unpack("<q", raw_record[8:16])[0] 
        record['seq'] = struct.unpack("<H", raw_record[16:18])[0]
        record['link'] = struct.unpack("<H", raw_record[18:20])[0]
        record['attr_off'] = struct.unpack("<H", raw_record[20:22])[0]
        record['flags'] = struct.unpack("<H", raw_record[22:24])[0]
        record['size'] = struct.unpack("<I", raw_record[24:28])[0]
        record['alloc_sizef'] = struct.unpack("<I", raw_record[28:32])[0]
        record['base_ref'] = struct.unpack("<q", raw_record[32:40])[0] 
        record['next_attrid'] = struct.unpack("<H", raw_record[40:42])[0]
        record['f1'] = raw_record[42:44]  
        record['recordnum'] = struct.unpack("<I", raw_record[44:48])[0]

        if record['base_ref'] & 0x8000000000000000:
            record['base_ref'] = -(~record['base_ref'] & 0xFFFFFFFFFFFFFFFF) - 1   

        
    def decodeATRHeader(self, s: bytes) -> Dict[str, Any]:

        d = {}

        d['type'] = struct.unpack("<L",s[:4])[0]

        if d['type'] == 0xffffffff:
            return d
        
        d['len']      = struct.unpack("<L",s[4:8])[0]
        d['res']      = struct.unpack( "B",s[8])[0]
        d['nlen']     = struct.unpack( "B",s[9])[0]
        d['name_off'] = struct.unpack("<H",s[10:12])[0]
        d['flags']    = struct.unpack("<H",s[12:14])[0]
        d['id']       = struct.unpack("<H",s[14:16])[0]

        if d['res'] == 0:
            d['ssize']   = struct.unpack("<L",s[16:20])[0]
            d['soff']    = struct.unpack("<H",s[20:22])[0]
            d['idxflag'] = struct.unpack("<H",s[22:24])[0]

        else:
            d['start_vcn'] = struct.unpack("<d",s[16:24])[0]
            d['last_vcn']  = struct.unpack("<d",s[24:32])[0]
            d['run_off']   = struct.unpack("<H",s[32:34])[0]
            d['compusize'] = struct.unpack("<H",s[34:36])[0]
            d['f1']        = struct.unpack("<I",s[36:40])[0]
            d['alen']      = struct.unpack("<d",s[40:48])[0]
            d['ssize']     = struct.unpack("<d",s[48:56])[0]
            d['initsize']  = struct.unpack("<d",s[56:64])[0]

        return d
    
    def handle_standard_information(self, ATRrecord: Dict[str, Any], raw_record: bytes, record: Dict[str, Any]) -> None:
        if self.options.debug:
            print(f"Standard Information:\n++Type: {hex(ATRrecord['type'])} Length: {ATRrecord['len']} Resident: {ATRrecord['res']} Name Len: {ATRrecord['nlen']} Name Offset: {ATRrecord['name_off']}")
        SIrecord = self.decodeSIAttribute(raw_record[ATRrecord['soff']:])
        record['si'] = SIrecord
        if self.options.debug:
            print(f"++CRTime: {SIrecord['crtime'].dtstr}\n++MTime: {SIrecord['mtime'].dtstr}\n++ATime: {SIrecord['atime'].dtstr}\n++EntryTime: {SIrecord['ctime'].dtstr}")

    def handle_attribute_list(self, ATRrecord: Dict[str, Any], raw_record: bytes, record: Dict[str, Any]) -> None:
        if self.options.debug:
            print("Attribute list")
        if ATRrecord['res'] == 0:
            ALrecord = self.decodeAttributeList(raw_record[ATRrecord['soff']:], record)
            record['al'] = ALrecord
            if self.options.debug:
                print(f"Name: {ALrecord['name']}")
        else:
            if self.options.debug:
                print("Non-resident Attribute List?")
            record['al'] = None

    def handle_file_name(self, ATRrecord: Dict[str, Any], raw_record: bytes, record: Dict[str, Any]) -> None:
        if self.options.debug:
            print("File name record")
        FNrecord = self.decodeFNAttribute(raw_record[ATRrecord['soff']:], record)
        record[('fn', record['fncnt'])] = FNrecord
        if self.options.debug:
            print(f"Name: {FNrecord['name']} ({record['fncnt']})")
        record['fncnt'] += 1
        if FNrecord['crtime'] != 0:
            if self.options.debug:
                print(f"\tCRTime: {FNrecord['crtime'].dtstr} MTime: {FNrecord['mtime'].dtstr} ATime: {FNrecord['atime'].dtstr} EntryTime: {FNrecord['ctime'].dtstr}")

    def handle_object_id(self, ATRrecord: Dict[str, Any], raw_record: bytes, record: Dict[str, Any]) -> None:
        ObjectIDRecord = self.decodeObjectID(raw_record[ATRrecord['soff']:])
        record['objid'] = ObjectIDRecord
        if self.options.debug:
            print("Object ID")

    def handle_security_descriptor(self, ATRrecord: Dict[str, Any], raw_record: bytes, record: Dict[str, Any]) -> None:
        record['sd'] = True
        if self.options.debug:
            print("Security descriptor")

    def handle_volume_name(self, ATRrecord: Dict[str, Any], raw_record: bytes, record: Dict[str, Any]) -> None:
        record['volname'] = True
        if self.options.debug:
            print("Volume name")

    def handle_volume_information(self, ATRrecord: Dict[str, Any], raw_record: bytes, record: Dict[str, Any]) -> None:
        if self.options.debug:
            print("Volume info attribute")
        VolumeInfoRecord = self.decodeVolumeInfo(raw_record[ATRrecord['soff']:])
        record['volinfo'] = VolumeInfoRecord

    def handle_data(self, ATRrecord: Dict[str, Any], raw_record: bytes, record: Dict[str, Any]) -> None:
        record['data'] = True
        if self.options.debug:
            print("Data attribute")

    def handle_index_root(self, ATRrecord: Dict[str, Any], raw_record: bytes, record: Dict[str, Any]) -> None:
        record['indexroot'] = True
        if self.options.debug:
            print("Index root")

    def handle_index_allocation(self, ATRrecord: Dict[str, Any], raw_record: bytes, record: Dict[str, Any]) -> None:
        record['indexallocation'] = True
        if self.options.debug:
            print("Index allocation")

    def handle_bitmap(self, ATRrecord: Dict[str, Any], raw_record: bytes, record: Dict[str, Any]) -> None:
        record['bitmap'] = True
        if self.options.debug:
            print("Bitmap")

    def handle_reparse_point(self, ATRrecord: Dict[str, Any], raw_record: bytes, record: Dict[str, Any]) -> None:
        record['reparsepoint'] = True
        if self.options.debug:
            print("Reparse point")

    def handle_ea_information(self, ATRrecord: Dict[str, Any], raw_record: bytes, record: Dict[str, Any]) -> None:
        record['eainfo'] = True
        if self.options.debug:
            print("EA Information")

    def handle_ea(self, ATRrecord: Dict[str, Any], raw_record: bytes, record: Dict[str, Any]) -> None:
        record['ea'] = True
        if self.options.debug:
            print("EA")

    def handle_property_set(self, ATRrecord: Dict[str, Any], raw_record: bytes, record: Dict[str, Any]) -> None:
        record['propertyset'] = True
        if self.options.debug:
            print("Property set")

    def handle_logged_utility_stream(self, ATRrecord: Dict[str, Any], raw_record: bytes, record: Dict[str, Any]) -> None:
        record['loggedutility'] = True
        if self.options.debug:
            print("Logged utility stream")

    def handle_unknown_attribute(self, ATRrecord: Dict[str, Any], raw_record: bytes, record: Dict[str, Any]) -> None:
        if self.options.debug:
            print(f"Found an unknown attribute type: {ATRrecord['type']:x}")

    def add_note(self, record: Dict[str, Any], note: str) -> None:
        if record['notes'] == '':
            record['notes'] = note
        else:
            record['notes'] += f" | {note} |"   
    
    def decodeSIAttribute(self, s: bytes) -> Dict[str, Any]:
        d = {}
        d['crtime'] = WindowsTime(struct.unpack("<L", s[:4])[0], struct.unpack("<L", s[4:8])[0], self.options.localtz)
        d['mtime'] = WindowsTime(struct.unpack("<L", s[8:12])[0], struct.unpack("<L", s[12:16])[0], self.options.localtz)
        d['ctime'] = WindowsTime(struct.unpack("<L", s[16:20])[0], struct.unpack("<L", s[20:24])[0], self.options.localtz)
        d['atime'] = WindowsTime(struct.unpack("<L", s[24:28])[0], struct.unpack("<L", s[28:32])[0], self.options.localtz)
        d['dos'] = struct.unpack("<I", s[32:36])[0]
        d['maxver'] = struct.unpack("<I", s[36:40])[0]
        d['ver'] = struct.unpack("<I", s[40:44])[0]
        d['class_id'] = struct.unpack("<I", s[44:48])[0]
        d['own_id'] = struct.unpack("<I", s[48:52])[0]
        d['sec_id'] = struct.unpack("<I", s[52:56])[0]
        d['quota'] = struct.unpack("<d", s[56:64])[0]
        d['usn'] = struct.unpack("<d", s[64:72])[0]
        return d

    def decodeFNAttribute(self, s: bytes, record: Dict[str, Any]) -> Dict[str, Any]:
        d = {}
        d['par_ref'] = struct.unpack("<Lxx", s[:6])[0]
        d['par_seq'] = struct.unpack("<H", s[6:8])[0]
        d['crtime'] = WindowsTime(struct.unpack("<L", s[8:12])[0], struct.unpack("<L", s[12:16])[0], self.options.localtz)
        d['mtime'] = WindowsTime(struct.unpack("<L", s[16:20])[0], struct.unpack("<L", s[20:24])[0], self.options.localtz)
        d['ctime'] = WindowsTime(struct.unpack("<L", s[24:28])[0], struct.unpack("<L", s[28:32])[0], self.options.localtz)
        d['atime'] = WindowsTime(struct.unpack("<L", s[32:36])[0], struct.unpack("<L", s[36:40])[0], self.options.localtz)
        d['alloc_fsize'] = struct.unpack("<q", s[40:48])[0]
        d['real_fsize'] = struct.unpack("<q", s[48:56])[0]
        d['flags'] = struct.unpack("<d", s[56:64])[0]
        d['nlen'] = struct.unpack("B", s[64])[0]
        d['nspace'] = struct.unpack("B", s[65])[0]

        d['name'] = self.decode_unicode(s[66:], d['nlen'])

        if any(ord(c) > 127 for c in d['name']):
            self.add_note(record, 'Filename contains non-ASCII characters')

        return d

    def decodeAttributeList(self, s: bytes, record: Dict[str, Any]) -> Dict[str, Any]:
        d = {}
        d['type'] = struct.unpack("<I", s[:4])[0]
        d['len'] = struct.unpack("<H", s[4:6])[0]
        d['nlen'] = struct.unpack("B", s[6])[0]
        d['f1'] = struct.unpack("B", s[7])[0]
        d['start_vcn'] = struct.unpack("<d", s[8:16])[0]
        d['file_ref'] = struct.unpack("<Lxx", s[16:22])[0]
        d['seq'] = struct.unpack("<H", s[22:24])[0]
        d['id'] = struct.unpack("<H", s[24:26])[0]

        d['name'] = self.decode_unicode(s[26:], d['nlen'])

        if any(ord(c) > 127 for c in d['name']):
            self.add_note(record, 'Attribute name contains non-ASCII characters')

        return d



from setuptools import setup, find_packages

with open('README.txt', 'r', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='analyzeMFT',
    version='2.1.1',
    author='Benjamin Cance',
    author_email='bjc@tdx.li',
    packages=find_packages(),
    url='http://github.com/rowingdude/analyzeMFT',
    license='LICENSE.txt',
    description='Analyze the $MFT from a NTFS filesystem.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    scripts=['analyzeMFT.py'],
    install_requires=[
        'typing;python_version<"3.5"',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: System :: Filesystems',
        'Topic :: Utilities',
    ],
    python_requires='>=3.6',
)


#!/usr/bin/env python

# Version 2.1.1
#
# Author: Benjamin Cance (bjc@tdx.li)
# Copyright Benjamin Cance 2024
#
# 2-Aug-24 
# - Updating to current PEP
#
# 12-Aug-24
# - Seperating out functionality into specific files
# - Adding json export 

import json
import logging
from typing import Dict, Any, List
from .mft_utils import decodeMFTmagic, decodeMFTisactive, decodeMFTrecordtype


def mft_to_csv(record: Dict[str, Any], ret_header: bool) -> List[str]:

    try:
        if ret_header:
            return [
            'Record Number', 'Good', 'Active', 'Record type',
            'Sequence Number', 'Parent File Rec. #', 'Parent File Rec. Seq. #',
            'Filename #1', 'Std Info Creation date', 'Std Info Modification date',
            'Std Info Access date', 'Std Info Entry date', 'FN Info Creation date',
            'FN Info Modification date', 'FN Info Access date', 'FN Info Entry date',
            'Object ID', 'Birth Volume ID', 'Birth Object ID', 'Birth Domain ID',
            'Filename #2', 'FN Info Creation date', 'FN Info Modify date',
            'FN Info Access date', 'FN Info Entry date', 'Filename #3', 'FN Info Creation date',
            'FN Info Modify date', 'FN Info Access date', 'FN Info Entry date', 'Filename #4',
            'FN Info Creation date', 'FN Info Modify date', 'FN Info Access date',
            'FN Info Entry date', 'Standard Information', 'Attribute List', 'Filename',
            'Object ID', 'Volume Name', 'Volume Info', 'Data', 'Index Root',
            'Index Allocation', 'Bitmap', 'Reparse Point', 'EA Information', 'EA',
            'Property Set', 'Logged Utility Stream', 'Log/Notes', 'STF FN Shift', 'uSec Zero'
            ]

        if 'baad' in record:
            return [str(record['recordnum']), "BAAD MFT Record"]

        csv_string = [
            record['recordnum'],
            decodeMFTmagic(record),
            decodeMFTisactive(record),
            decodeMFTrecordtype(record),
            str(record['seq'])
        ]

        if 'corrupt' in record:
            return csv_string + [str(record['recordnum']), "Corrupt", "Corrupt", "Corrupt MFT Record"]

        tmp_string = ["%d" % record['seq']]
        csv_string.extend(tmp_string)

        if record['fncnt'] > 0:
            csv_string.extend([str(record['fn',0]['par_ref']), str(record['fn',0]['par_seq'])])
        else:
            csv_string.extend(['NoParent', 'NoParent'])

        if record['fncnt'] > 0 and 'si' in record:
            
            filenameBuffer = [record['filename'],  str(record['si']['crtime'].dtstr),
                    record['si']['mtime'].dtstr,    record['si']['atime'].dtstr, record['si']['ctime'].dtstr,
                    record['fn',0]['crtime'].dtstr, record['fn',0]['mtime'].dtstr,
                    record['fn',0]['atime'].dtstr,  record['fn',0]['ctime'].dtstr]
        elif 'si' in record:

            filenameBuffer = ['NoFNRecord', str(record['si']['crtime'].dtstr),
                    record['si']['mtime'].dtstr, record['si']['atime'].dtstr, record['si']['ctime'].dtstr,
                    'NoFNRecord', 'NoFNRecord', 'NoFNRecord','NoFNRecord']
        else:

            filenameBuffer = ['NoFNRecord', 'NoSIRecord', 'NoSIRecord', 'NoSIRecord', 'NoSIRecord',
                            'NoFNRecord', 'NoFNRecord', 'NoFNRecord', 'NoFNRecord']


        csv_string.extend(filenameBuffer)

        if 'objid' in record:
            objidBuffer = [record['objid']['objid'], record['objid']['orig_volid'],
                        record['objid']['orig_objid'], record['objid']['orig_domid']]
        else:
            objidBuffer = ['','','','']

        csv_string.extend(objidBuffer)

        for i in range(1, record['fncnt']):
            filenameBuffer = [record['fn',i]['name'], record['fn',i]['crtime'].dtstr, record['fn',i]['mtime'].dtstr,
                    record['fn',i]['atime'].dtstr, record['fn',i]['ctime'].dtstr]
            csv_string.extend(filenameBuffer)
            filenameBuffer = ''

        if record['fncnt'] < 2:
            tmp_string = ['','','','','','','','','','','','','','','']
        elif record['fncnt'] == 2:
            tmp_string = ['','','','','','','','','','']
        elif record['fncnt'] == 3:
            tmp_string = ['','','','','']

        csv_string.extend(tmp_string)

        
        attributes = ['si', 'al', 'objid', 'volname', 'volinfo', 'data', 'indexroot', 
                    'indexallocation', 'bitmap', 'reparse', 'eainfo', 'ea', 
                    'propertyset', 'loggedutility']

        csv_string.extend(
            ['True' if attr in record else 'False' for attr in attributes]
        )

        csv_string.append('True' if record.get('fncnt', 0) > 0 else 'False')

        if 'notes' in record:                        # Log of abnormal activity related to this record
            csv_string.append(record['notes'])
        else:
            csv_string.append('None')
            record['notes'] = ''

        if 'stf-fn-shift' in record:
            csv_string.append('Y')
        else:
            csv_string.append('N')

        if 'usec-zero' in record:
            csv_string.append('Y')
        else:
            csv_string.append('N')

        return csv_string
    
    except KeyError as e:

        logging.error(f"Missing key in record: {e}")
        return []

def mft_to_body(record: Dict[str, Any], full: bool, std: bool) -> str:
    try:

        if record['fncnt'] > 0:

            if full == True: # Use full path
                name = record['filename']
            else:
                name = record['fn',0]['name']

            if std == True:     # Use STD_INFO
                rec_bodyfile = ("%s|%s|%s|%s|%s|%s|%s|%d|%d|%d|%d\n" %
                            ('0',name,'0','0','0','0',
                            int(record['fn',0]['real_fsize']),
                            int(record['si']['atime'].unixtime),  # was str ....
                            int(record['si']['mtime'].unixtime),
                            int(record['si']['ctime'].unixtime),
                            int(record['si']['ctime'].unixtime)))
            else:               # Use FN
                rec_bodyfile = ("%s|%s|%s|%s|%s|%s|%s|%d|%d|%d|%d\n" %
                            ('0',name,'0','0','0','0',
                            int(record['fn',0]['real_fsize']),
                            int(record['fn',0]['atime'].unixtime),
                            int(record['fn',0]['mtime'].unixtime),
                            int(record['fn',0]['ctime'].unixtime),
                            int(record['fn',0]['crtime'].unixtime)))

        else:
            if 'si' in record:
                rec_bodyfile = ("%s|%s|%s|%s|%s|%s|%s|%d|%d|%d|%d\n" %
                            ('0','No FN Record','0','0','0','0', '0',
                            int(record['si']['atime'].unixtime),  # was str ....
                            int(record['si']['mtime'].unixtime),
                            int(record['si']['ctime'].unixtime),
                            int(record['si']['ctime'].unixtime)))
            else:
                rec_bodyfile = ("%s|%s|%s|%s|%s|%s|%s|%d|%d|%d|%d\n" %
                                    ('0','Corrupt Record','0','0','0','0', '0',0, 0, 0, 0))

        return rec_bodyfile
    
    except KeyError as e:

        logging.error(f"Missing key in record: {e}")
        return ""

def mft_to_l2t(record: Dict[str, Any]) -> str:
    try:
    
        if record['fncnt'] > 0:
            for i in ('atime', 'mtime', 'ctime', 'crtime'):
                (date,time) = record['fn',0][i].dtstr.split(' ')

                if i == 'atime':
                    type_str = '$FN [.A..] time'
                    macb_str = '.A..'
                if i == 'mtime':
                    type_str = '$FN [M...] time'
                    macb_str = 'M...'
                if i == 'ctime':
                    type_str = '$FN [..C.] time'
                    macb_str = '..C.'
                if i == 'crtime':
                    type_str = '$FN [...B] time'
                    macb_str = '...B'

                csv_string = ("%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s\n" %
                    (date, time, 'TZ', macb_str, 'FILE', 'NTFS $MFT', type_str, 'user', 'host', record['filename'], 'desc',
                    'version', record['filename'], record['seq'], record['notes'], 'format', 'extra'))

        elif 'si' in record:
            for i in ('atime', 'mtime', 'ctime', 'crtime'):
                (date,time) = record['si'][i].dtstr.split(' ')

                if i == 'atime':
                    type_str = '$SI [.A..] time'
                    macb_str = '.A..'
                if i == 'mtime':
                    type_str = '$SI [M...] time'
                    macb_str = 'M...'
                if i == 'ctime':
                    type_str = '$SI [..C.] time'
                    macb_str = '..C.'
                if i == 'crtime':
                    type_str = '$SI [...B] time'
                    macb_str = '...B'

                csv_string = ("%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s\n" %
                    (date, time, 'TZ', macb_str, 'FILE', 'NTFS $MFT', type_str, 'user', 'host', record['filename'], 'desc',
                    'version', record['filename'], record['seq'], record['notes'], 'format', 'extra'))

        else:
            csv_string = ("%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s|%s\n" %
                    ('-', '-', 'TZ', 'unknown time', 'FILE', 'NTFS $MFT', 'unknown time', 'user', 'host', 'Corrupt Record', 'desc',
                    'version', 'NoFNRecord', record['seq'], '-', 'format', 'extra'))

        return csv_string
    
    except KeyError as e:
        logging.error(f"Missing key in record: {e}")
        return ""
    
def mft_to_json(record: Dict[str, Any]) -> str:

    try:

        for key, value in record.items():
            if isinstance(value, dict):
                for sub_key, sub_value in value.items():
                    if hasattr(sub_value, 'dtstr'):
                        value[sub_key] = sub_value.dtstr
        return json.dumps(record, indent=2)
    
    except (TypeError, ValueError) as e:

        logging.error(f"Error converting record to JSON: {e}")
        return "{}"